@@ -21,6 +21,9 @@ import cffi
 import numpy as np
 
 
+__version__ = "1.6.0"
+
+
 def string_from_ffi(s):
     return ffi.string(s).decode("utf-8")
 
@@ -100,12 +103,14 @@ class MetviewInvoker:
             raise Exception(
                 'Command "metview" did not respond within '
                 + str(self.metview_startup_timeout)
-                + " seconds. "
+                + " seconds. This timeout is configurable by setting "
+                "environment variable METVIEW_PYTHON_START_TIMEOUT in seconds. "
                 "At least Metview 5 is required, so please ensure it is in your PATH, "
                 "as earlier versions will not work with the Python interface."
             )
 
         self.read_metview_settings(env_file.name)
+        env_file.close()
 
         # when the Python session terminates, we should destroy this object so that the Metview
         # session is properly cleaned up. We can also do this in a __del__ function, but there can
@@ -264,7 +269,7 @@ class Request(dict, Value):
                 self.verb = req.verb
                 self.val_pointer = req.val_pointer
             else:
-                if myverb != None:
+                if myverb is not None:
                     self.set_verb(myverb)
 
         # initialise from a Macro pointer
@@ -274,10 +279,7 @@ class Request(dict, Value):
             n = lib.p_get_req_num_params(req)
             for i in range(0, n):
                 param = string_from_ffi(lib.p_get_req_param(req, i))
-                raw_val = lib.p_get_req_value(req, param.encode("utf-8"))
-                if raw_val != ffi.NULL:
-                    val = string_from_ffi(raw_val)
-                    self[param] = val
+                self[param] = self[param]
             # self['_MACRO'] = 'BLANK'
             # self['_PATH']  = 'BLANK'
 
@@ -286,9 +288,10 @@ class Request(dict, Value):
 
     # translate Python classes into Metview ones where needed
     def to_metview_style(self):
-        for k, v in self.items():
+        for k in list(self):
 
             # bool -> on/off
+            v = self.get(k)  # self[k] returns 1 for True
             if isinstance(v, bool):
                 conversion_dict = {True: "on", False: "off"}
                 self[k] = conversion_dict[v]
@@ -379,14 +382,17 @@ def push_vector(npa):
         f32_array = npa.astype(np.float32)
         cffi_buffer = ffi.cast("float*", f32_array.ctypes.data)
         lib.p_push_vector_from_float32_array(cffi_buffer, len(f32_array), np.nan)
+    elif dtype == np.int:  # convert first to float64
+        f64_array = npa.astype(np.float64)
+        cffi_buffer = ffi.cast("double*", f64_array.ctypes.data)
+        lib.p_push_vector_from_double_array(cffi_buffer, len(f64_array), np.nan)
     else:
-        raise Exception(
+        raise TypeError(
             "Only float32 and float64 numPy arrays can be passed to Metview, not ",
             npa.dtype,
         )
 
 
-
 class File(Value):
     def __init__(self, val_pointer):
         Value.__init__(self, val_pointer)
@@ -396,6 +402,9 @@ class FileBackedValue(Value):
     def __init__(self, val_pointer):
         Value.__init__(self, val_pointer)
 
+    def write(self, filename):
+        return write(filename, self)
+
     def url(self):
         # ask Metview for the file relating to this data (Metview will write it if necessary)
         return string_from_ffi(lib.p_data_path(self.val_pointer))
@@ -462,6 +471,15 @@ class FileBackedValueWithOperators(FileBackedValue):
     def __abs__(self):
         return self.abs()
 
+    def __and__(self, other):
+        return met_and(self, other)
+
+    def __or__(self, other):
+        return met_or(self, other)
+
+    def __invert__(self):
+        return met_not(self)
+
 
 class ContainerValueIterator:
     def __init__(self, data):
@@ -511,12 +529,14 @@ class ContainerValue(Value):
                         f = merge(f, fields[i])
                     return f
             else:
-                raise Exception(
+                raise IndexError(
                     "This object does not support extended slicing: " + str(self)
                 )
         else:  # normal index
             if isinstance(index, str):  # can have a string as an index
                 return subset(self, index)
+            elif isinstance(index, np.ndarray):  # can have an array as an index
+                return subset(self, index + self.macro_index_base)
             else:
                 if index < 0:  # negative index valid for range [-len..-1]
                     c = int(count(self))
@@ -536,14 +556,14 @@ class ContainerValue(Value):
             push_arg(value)
             lib.p_set_subvalue_from_arg_stack(self.val_pointer)
         else:
-            raise Exception("Cannot assign ", value, " as element of ", self)
+            raise IndexError("Cannot assign ", value, " as element of ", self)
 
     def __iter__(self):
         return ContainerValueIterator(self)
 
 
 class Fieldset(FileBackedValueWithOperators, ContainerValue):
-    def __init__(self, val_pointer=None, path=None):
+    def __init__(self, val_pointer=None, path=None, fields=None):
         FileBackedValueWithOperators.__init__(self, val_pointer)
         ContainerValue.__init__(
             self,
@@ -553,10 +573,17 @@ class Fieldset(FileBackedValueWithOperators, ContainerValue):
             support_slicing=True,
         )
 
+        if (path is not None) and (fields is not None):
+            raise ValueError("Fieldset cannot take both path and fields")
+
         if path is not None:
             temp = read(path)
             self.steal_val_pointer(temp)
 
+        if fields is not None:
+            for f in fields:
+                self.append(f)
+
     def append(self, other):
         temp = merge(self, other)
         if self.val_pointer != None:  #  we will overwrite ourselves, so delete
@@ -980,11 +1007,16 @@ def bind_functions(namespace, module_name=None):
     # HACK: some fuctions are missing from the 'dictionary' call.
     namespace["neg"] = make("neg")
     namespace["nil"] = make("nil")
+    namespace["div"] = div
+    namespace["mod"] = mod
     # override some functions that need special treatment
     # FIXME: this needs to be more structured
     namespace["plot"] = plot
     namespace["setoutput"] = setoutput
+    namespace["metzoom"] = metzoom
+    namespace["version_info"] = version_info
     namespace["dataset_to_fieldset"] = dataset_to_fieldset
+    namespace["download_gallery_data"] = download_gallery_data
 
     namespace["Fieldset"] = Fieldset
     namespace["Request"] = Request
@@ -995,6 +1027,7 @@ add = make("+")
 call = make("call")
 count = make("count")
 div = make("/")
+download = make("download")
 equal = make("=")
 filter = make("filter")
 greater_equal_than = make(">=")
@@ -1004,6 +1037,7 @@ lower_than = make("<")
 merge = make("&")
 met_not_eq = make("<>")
 met_plot = make("plot")
+mod = make("mod")
 nil = make("nil")
 png_output = make("png_output")
 power = make("^")
@@ -1011,8 +1045,22 @@ prod = make("*")
 ps_output = make("ps_output")
 read = make("read")
 met_setoutput = make("setoutput")
+metzoom = make("metzoom")
 sub = make("-")
 subset = make("[]")
+met_and = make("and")
+met_or = make("or")
+met_not = make("not")
+met_version_info = make("version_info")
+write = make("write")
+
+
+# call the C++ version_info() function and add the version of the
+# Python bindings to the resulting dict
+def version_info():
+    binary_info = dict(met_version_info())
+    binary_info['metview_python_version'] = __version__
+    return binary_info
 
 
 # -----------------------------------------------------------------------------
@@ -1023,6 +1071,7 @@ subset = make("[]")
 class Plot:
     def __init__(self):
         self.plot_to_jupyter = False
+        self.jupyter_args = {}
 
     def __call__(self, *args, **kwargs):
         if self.plot_to_jupyter:
@@ -1031,8 +1080,9 @@ class Plot:
 
             base, ext = os.path.splitext(tmp)
 
+            self.jupyter_args.update(output_name=base, output_name_first_page_number="off")
             met_setoutput(
-                png_output(output_name=base, output_name_first_page_number="off")
+                png_output(self.jupyter_args)
             )
             met_plot(*args)
 
@@ -1062,7 +1112,7 @@ plot = Plot()
 # under most circumstances, we only import it when the user asks for Jupyter
 # functionality. Since this occurs within a function, we need a little trickery to
 # get the IPython functions into the global namespace so that the plot object can use them
-def setoutput(*args):
+def setoutput(*args, **kwargs):
     if "jupyter" in args:
         try:
             global Image
@@ -1077,6 +1127,7 @@ def setoutput(*args):
         # test whether we're in the Jupyter environment
         if get_ipython() is not None:
             plot.plot_to_jupyter = True
+            plot.jupyter_args = kwargs
         else:
             print(
                 "ERROR: setoutput('jupyter') was set, but we are not in a Jupyter environment"
@@ -1085,3 +1136,16 @@ def setoutput(*args):
     else:
         plot.plot_to_jupyter = False
         met_setoutput(*args)
+
+
+# -----------------------------------------------------------------------------
+#                        Other user-visible utiliy functions
+# -----------------------------------------------------------------------------
+
+def download_gallery_data(filename):
+    base_url = "http://download.ecmwf.org/test-data/metview/gallery/"
+    try:
+        d = download(url=base_url + filename, target=filename)
+        return d
+    except:
+        raise Exception("Could not download file " + filename + " from the download server")
